# -*- coding: UTF-8 -*-
# doctest rstgen/utils.py
# Copyright 2009-2021 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

"""
Defines a series of utility classes and functions.

"""

import re
import os
import sys
import datetime
import inspect
import subprocess
from dateutil import parser as dateparser
from contextlib import contextmanager
from pathlib import Path
from pprint import pprint
from importlib import import_module


def confirm(prompt=None, default="y"):
    """
    Ask for user confirmation from the console.
    """
    # if six.PY2:
    #     prompt = prompt.encode(
    #         sys.stdin.encoding or locale.getpreferredencoding(True))
    # print(20160324, type(prompt))
    prompt += " [Y,n]?"
    while True:
        ln = input(prompt)
        if not ln:
            ln = default
        if ln.lower() in ('y', 'j', 'o'):
            return True
        if ln.lower() == 'n':
            return False
        print("Please answer Y or N")

def i2d(i):
    """
    Convert `int` to `date`. Examples:

    >>> i2d(20121224)
    datetime.date(2012, 12, 24)

    """
    s = str(i)
    if len(s) != 8:
        raise Exception("Invalid date specification {0}.".format(i))
    d = dateparser.parse(s)
    d = datetime.date(d.year, d.month, d.day)
    # print(i, "->", v)
    return d


def i2t(s):
    """
    Convert `int` to `time`. Examples:

    >>> i2t(815)
    datetime.time(8, 15)

    >>> i2t(1230)
    datetime.time(12, 30)

    >>> i2t(12)
    datetime.time(12, 0)

    >>> i2t(1)
    datetime.time(1, 0)

    """
    s = str(s)
    if len(s) == 4:
        return datetime.time(int(s[:2]), int(s[2:]))
    if len(s) == 3:
        return datetime.time(int(s[:1]), int(s[1:]))
    if len(s) <= 2:
        return datetime.time(int(s), 0)
    raise ValueError(s)


def indentation(s):
    """
    Examples:

    >>> indentation("")
    0
    >>> indentation("foo")
    0
    >>> indentation(" foo")
    1

    """
    return len(s) - len(s.lstrip())


def unindent(s):
    """
    Reduces indentation of a docstring to the minimum.
    Empty lines don't count.

    Examples:

    >>> unindent('')
    ''
    >>> print(unindent('''
    ...   foo
    ...     foo
    ... '''))
    <BLANKLINE>
    foo
      foo
    >>> print(unindent('''
    ... foo
    ...     foo
    ... '''))
    <BLANKLINE>
    foo
        foo
    """
    s = s.rstrip()
    lines = s.splitlines()
    if len(lines) == 0:
        return s.strip()
    mini = sys.maxsize
    for ln in lines:
        ln = ln.rstrip()
        if len(ln) > 0:
            mini = min(mini, indentation(ln))
            if mini == 0:
                break
    if mini == sys.maxsize:
        return s
    return '\n'.join([i[mini:] for i in lines])


class SubProcessParent(object):

    """
    Base class for :class:`atelier.test.TestCase`.
    Also used standalone by `lino.management.commands.makescreenshots`.
    """
    default_environ = dict()
    # inheritable_envvars = ('VIRTUAL_ENV', 'PYTHONPATH', 'PATH')

    def build_environment(self):
        """Contructs and return a `dict` with the environment variables for
        the future subprocess.

        """
        env = dict()
        env.update(os.environ)
        env.update(self.default_environ)
        # env.update(COVERAGE_PROCESS_START="folder/.coveragerc")
        # for k in self.inheritable_envvars:
        #     v = os.environ.get(k, None)
        #     if v is not None:
        #         env[k] = v
        return env

    def check_output(self, args, **kw):
        env = self.build_environment()
        kw.update(env=env)
        kw.update(stderr=subprocess.STDOUT)
        return subprocess.check_output(args, **kw)

    def open_subprocess(self, args, **kw):
        """Additional keywords will be passed to the `Popen constructor
        <http://docs.python.org/2.7/library/subprocess.html#popen-constructor>`_.
        They can be e.g.  `cwd` : the working directory

        """
        env = self.build_environment()
        # raise Exception("20170912 {}".format(env.keys()))
        kw.update(env=env)
        #~ subprocess.check_output(args,**kw)
        #~ from StringIO import StringIO
        #~ buffer = StringIO()
        kw.update(stdout=subprocess.PIPE)
        kw.update(stderr=subprocess.STDOUT)
        kw.update(universal_newlines=True)
        return subprocess.Popen(args, **kw)

    def run_subprocess(self, args, **kw):
        """
        Run a subprocess, wait until it terminates,
        fail if the returncode is not 0.
        """
        # print ("20150214 run_subprocess %r" % args)
        p = self.open_subprocess(args, **kw)

        # wait() will deadlock when using stdout=PIPE and/or
        # stderr=PIPE and the child process generates enough output to
        # a pipe such that it blocks waiting for the OS pipe buffer to
        # accept more data. Use communicate() to avoid that.
        if False:
            p.wait()
        else:
            out, err = p.communicate()
        # raise Exception("20160711 run_subprocess", out)
        rv = p.returncode
        # kw.update(stderr=buffer)
        # rv = subprocess.call(args,**kw)
        if rv != 0:
            cmd = ' '.join(map(str, args))  # args can contain PosixPath instances
            # if six.PY2:
            #     # if the output contains non-asci chars, then we must
            #     # decode here in order to wrap it into our msg. Later
            #     # we must re-encode it because exceptions, in Python
            #     # 2, don't want unicode strings.
            #     out = out.decode("utf-8")
            msg = "%s (%s) returned %d:\n-----\n%s\n-----" % (
                cmd, kw, rv, out)
            # try:
            #     msg = "%s (%s) returned %d:\n-----\n%s\n-----" % (
            #         cmd, kw, rv, out)
            # except UnicodeDecodeError:
            #     out = repr(out)
            #     msg = "%s (%s) returned %d:OOPS\n-----\n%s\n-----" % (
            #         cmd, kw, rv, out)

            # print msg
            # if six.PY2:
            #     msg = msg.encode('utf-8')
            self.fail(msg)


@contextmanager
def cd(path):
    old_dir = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(old_dir)


def srcref_url_template(mod):
    root_mod_name = mod.__name__.split('.')[0]
    root_mod = __import__(root_mod_name)
    setup_info = getattr(root_mod, 'SETUP_INFO', {})
    url = setup_info.get('url')
    if url is None:
        return (root_mod, None)
    if not url.endswith('/'):
        url += "/"
    # srcref_url = getattr(root_mod, 'srcref_url', None)
    # if srcref_url is None:
    #     # print(20180126, root_module_name, root_mod, srcref_url)
    #     return
    #~ if not mod.__name__.startswith('lino.'):
        #~ return
    return (root_mod, url + "blob/master/%s")

def srcref(mod):
    root_mod, tpl = srcref_url_template(mod)
    if tpl is None:
        return
    srcref = mod.__file__
    if srcref.endswith('.pyc'):
        srcref = srcref[:-1]
    if True:
        # failed on readthedocs.org because there was a dangling pyc
        # file on my machine which caused autodoc to create an entry
        # in docs/api.
        if os.stat(srcref).st_size == 0:
            return
    #~ srcref = srcref[len(lino.__file__)-17:]
    root = str(Path(root_mod.__file__).parents[1])
    if len(root):
        srcref = srcref[len(root) + 1:]
    srcref = srcref.replace(os.path.sep, '/')
    return tpl % srcref


def import_from_dotted_path(dotted_names, path=None):
    """
    Thanks to Chase Seibert,
    https://chase-seibert.github.io/blog/2014/04/23/python-imp-examples.html
    """

    s = dotted_names.split('.', 1)
    if len(s) == 2:
        first, remaining = s
    else:
        first, remaining = dotted_names, None
    fp, pathname, description = importlib.find_module(first, path)
    module = importlib.load_module(first, fp, pathname, description)
    if not remaining:
        return (module, None)
    if hasattr(module, remaining):
        return (module, getattr(module, remaining))
    return import_from_dotted_path(remaining, path=module.__path__)


def py2url_txt(s):
    """
    Return a tuple `(url, txt)` where `url` is the URL which links to
    the source code of the specified Python object and `txt` is the
    suggested short text to use in a hyperlink.
    """
    args = s.split(None, 1)
    if len(args) == 1:
        txt = s
    else:
        s = args[0]
        txt = args[1]

    if False:
        try:
            mod, obj = import_from_dotted_path(s)
            return (srcref(mod), txt)
        except Exception as e:
            return ("Error in Python code ({})".format(e), txt)
    parts = s.split('.')
    try:
        obj = import_module(parts[0])
        for p in parts[1:]:
            obj = getattr(obj, p)
        mod = inspect.getmodule(obj)
        return (srcref(mod), txt)
    except Exception as e:
        return ("Error in Python code ({})".format(e), txt)


def dict_py2(old_dict):
    """Convert the given `dict` so that it's `repr` is the same for both
    Python 2 and 3.

    Deprecated. Use :func:`rmu` instead.
    """
    from future.utils import viewitems
    new_dict = {}
    for (key, value) in viewitems(old_dict):
        if type(value) == dict:
            new_dict[str(key)] = dict_py2(value)
        elif type(value) == list:
            new_dict[str(key)] = list_py2(value)
        elif type(value) == tuple:
            new_dict[str(key)] = tuple_py2(value)
        else:
            if isinstance(value, bool):
                new_dict[str(key)] = value
            else:
                new_dict[str(key)] = str(value)
    return new_dict


def list_py2(old_list):
    """Convert the given `list` so that it's `repr` is the same for both
    Python 2 and 3.

    Deprecated. Use :func:`rmu` instead.

    """
    new_list = []
    for item in old_list:
        if type(item) == dict:
            new_list.append(dict_py2(item))
        elif type(item) == tuple:
            new_list.append(tuple_py2(item))
        else:
            new_list.append(str(item))
    return new_list


def tuple_py2(old_tuple):
    """Convert the given `tuple` so that it's `repr` is the same for both
    Python 2 and 3.

    Deprecated. Use :func:`rmu` instead.

    """
    lst = list(old_tuple)
    lst = list_py2(lst)
    return tuple(lst)


def rmu(x):
    """Remove the 'u' prefix from unicode strings under Python 2 in order
    to produce Python 3 compatible output in a doctested code snippet.

    >>> lst = [123, "123", u"Äöü"]
    >>> print(rmu(lst))
    [123, '123', '\\xc4\\xf6\\xfc']
    >>> print(rmu(tuple(lst)))
    (123, '123', '\\xc4\\xf6\\xfc')
    >>> dct = {i: i for i in lst}
    >>> print(rmu(dct)) #doctest: +ELLIPSIS
    {...'\\xc4\\xf6\\xfc': '\\xc4\\xf6\\xfc'...}

    """
    if isinstance(x, Path):
        return x
    # if isinstance(x, collections.namedtuple):
    #     return x
    if isinstance(x, list):
        return [rmu(i) for i in x]
    if isinstance(x, tuple):
        return tuple([rmu(i) for i in x])
    if isinstance(x, dict):
        return {rmu(k):rmu(v) for k,v in x.items()}
    if isinstance(x, str):
        return str(x)
    return x


def sixprint(*args):
    """Like print, but simulating PY3 output under PY2."""
    for x in args:
        # if six.PY2 and isinstance(x, set):
        #     print("{%s}" % ', '.join([str(rmu(i)) for i in x]))
        # else:
        pprint(rmu(x))



# -*- coding: utf-8 -*-
# Copyright 2011-2022 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

"""

Basic extension

Sphinx setup used to build the Lino documentation.


.. rst:role:: blogref

    Inserts a reference to the blog entry of the specified date.

    Instead of writing ``:doc:`/blog/2011/0406``` it is better to
    write ``:blogref:`20110406``` because the latter works between
    Sphinx trees and also supports archived blog entries.

Thanks to

- `Creating reStructuredText Directives
  <http://docutils.sourceforge.net/docs/howto/rst-directives.html>`_

"""

import os
import inspect
import importlib
from pathlib import Path

from docutils import nodes, utils
from docutils.parsers.rst import directives
from docutils.parsers.rst import roles
# from docutils.parsers.rst import Directive
#from sphinx.util.compat import Directive
from sphinx.util.nodes import nested_parse_with_titles
from sphinx.util.nodes import split_explicit_title
from sphinx import addnodes
from sphinx.util.docutils import SphinxRole

from importlib import import_module
from rstgen.utils import i2d, srcref, import_from_dotted_path, py2url_txt
import atelier

# from docutils.languages import LanguageImporter as li
# if li.packages == ('docutils.languages.', ''):
#     li.packages = ('docutils.languages.', 'atelier.sphinxcontrib.languages', '')

def autodoc_skip_member(app, what, name, obj, skip, options):
    defmod = getattr(obj, '__module__', None)
    if defmod is not None:
        try:
            if defmod.startswith('django.'):
                return True
        except AttributeError as e:
            # raise Exception("{!r} : {}".format(obj, e))
            pass
    if name == 'Q':
        print(20141219, app, what, name, obj, skip, options)
    if name != '__builtins__':
        #~ print 'autodoc_skip_member', what, repr(name), repr(obj)

        if what == 'class':
            if name.endswith('MultipleObjectsReturned'):
                return True
            if name.endswith('DoesNotExist'):
                return True

            #~ if isinstance(obj,ObjectDoesNotExist) \
              #~ or isinstance(obj,MultipleObjectsReturned):
                #~ return True

    #~ if what == 'exception':
        #~ print 'autodoc_skip_member',what, repr(name), repr(obj), skip
        #~ return True

    return skip


SRCREF_TEMPLATE = """

(This module's source code is available `here <%s>`__.)

"""

SRCREF_TEMPLATE_BEFORE = """

.. note:: This module's source code is available
   `here <%s>`__.

"""

#~ """
#~ SIDEBAR = """
#~ (source code: :srcref:`/%s`)

#~ """

SIDEBAR = """
(`source code <%s>`__)

"""

#~ SIDEBAR = """
#~ .. sidebar:: Use the source, Luke

  #~ We generally recommend to also consult the source code.
  #~ This module's source code is available at
  #~ :srcref:`/%s.py`

#~ """


def process_signature(app, what, name, obj, options, signature,
                      return_annotation):
    # experimental. not yet used.
    # trying to get source links à la django doc.
    # test it in atelier with `$ inv clean -b bd`
    # currently this gives
    # Exception occurred:
    #   File ".../atelier/setup.py", line 3, in <module>
    #     exec(compile(fd.read(), fn, 'exec'))
    # IOError: [Errno 2] No such file or directory: 'atelier/setup_info.py'
    # if signature or return_annotation:
    #     raise Exception(
    #         "20170118 {!r} {!r} {!r} {!r} {!r} {!r}".format(
    #             what, name, obj, options, signature,
    #             return_annotation))
    if what == 'module':
        assert not signature
        s = srcref(obj)
        if s:
            signature = " [{}]".format(s)
        # signature = " [foo]"
        print(20170118, signature)
    return (signature, return_annotation)

def autodoc_add_srcref(app, what, name, obj, options, lines):
    """Add a reference to the module's source code.
    This is being added as listener to the
    `autodoc-process-docstring <http://sphinx-doc.org/ext/autodoc.html#event-autodoc-process-docstring>`_ signal.

    """
    if what == 'module':
        s = srcref(obj)
        if s:
            # We must add it *after* the module description (not
            # before) because also autosummary gets the docstring
            # processed by this handler, and the overview table in
            # the parent module would show always that same sentence.
            # I tried whether autosummary is intelligent and removes
            # admonitions when generating the summary: unfortunately
            # not.
            # 20151006 app.env.config.html_context.update(source_code_link=s)
            if False: # 20220126 deactivate srcref mention (for Sphinx 4.4)
                if True:
                    lines += (SRCREF_TEMPLATE % s).splitlines()
                else:
                    s = (SRCREF_TEMPLATE_BEFORE % s).splitlines()
                    # s = (SIDEBAR % s).splitlines()
                    s.reverse()
                    for ln in s:
                        lines.insert(0, ln)
                    # print('\n'.join(lines))


def get_blog_url(env, today):
    """
    Return the URL to your developer blog entry of that date.
    """
    # if today.year < 2013:  # TODO: make this configurable
    #     blogger_project = "lino"
    #     url_root = "http://code.google.com/p/%s/source/browse/" % blogger_project
    #     parts = ('docs', 'blog', str(today.year), today.strftime("%m%d.rst"))
    #     return url_root + "/".join(parts)

    fmt = env.config.blogref_format
    if not fmt:
        # return "oops"
        msg = "Please set your `blogref_format` to something "
        msg = "like 'https://www.example.com/blog/%Y/%m%d.html'."
        # msg += "\n(settings.keys are %s)" % env.settings.keys()
        raise Exception(msg)
    url = today.strftime(fmt)
    return url


def blogref_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    # thanks to http://docutils.sourceforge.net/docs/howto/rst-roles.html
    # this code originally from roles.pep_reference_role
    #~ print 20130315, rawtext, text, utils.unescape(text)
    has_explicit_title, title, target = split_explicit_title(text)
    try:
        date = i2d(int(target))
    except ValueError:
        msg = inliner.reporter.error(
            'Invalid text %r: must be an integer date of style "20130315" .'
            % text, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    #~ print repr(env)
    #~ raise Exception(20130315)
    #~ ref = inliner.document.settings.pep_base_url
           #~ + inliner.document.settings.pep_file_url_template % date)
    roles.set_classes(options)
    #~ from django.conf import settings
    #~ shown_text = settings.SITE.dtos(date)
    env = inliner.document.settings.env
    if not has_explicit_title:
        title = date.strftime(env.settings.get('today_fmt', '%Y-%m-%d'))
    title = utils.unescape(title)
    return [nodes.reference(rawtext, title,
                            refuri=get_blog_url(env, date),
                            **options)], []


def message_role(typ, rawtext, text, lineno, inliner, options={}, content=[]):
    text = utils.unescape(text)
    has_explicit_title, title, target = split_explicit_title(text)
    node = nodes.literal(rawtext, text)
    return [node], []

def actor_role(typ, rawtext, text, lineno, inliner, options={}, content=[]):
    text = utils.unescape(text)
    has_explicit_title, title, target = split_explicit_title(text)
    node = nodes.literal(rawtext, text)
    return [node], []

def coderef_role(typ, rawtext, text, lineno, inliner, options={}, content=[]):
    text = utils.unescape(text)
    has_explicit_title, title, target = split_explicit_title(text)
    try:
        modname, name = target.rsplit('.', 1)
    except ValueError:
        raise Exception("Don't know how to import name %s" % target)
    mod = import_module(modname)

    try:
        value = getattr(mod, name, None)
    except AttributeError:
        raise Exception("No name '%s' in module '%s'" % (name, modname))
    #~ raise Exception("20130908 %s " % lines)
    if isinstance(value, type):
        if value.__module__ != modname:
            raise Exception("20130908 %r != %r" % (value.__module__, modname))

    url = srcref(mod)

    lines, line_no = inspect.getsourcelines(value)
    if line_no:
        url += "#" + str(line_no)
    if not has_explicit_title:
        pass
    pnode = nodes.reference(title, title, internal=False, refuri=url)
    return [pnode], []


def unused_srcref_role(typ, rawtext, text, lineno, inliner, options={}, content=[]):
    text = utils.unescape(text)
    has_explicit_title, title, target = split_explicit_title(text)
    url = srcref(target)
    try:
        full_url = base_url % part
    except (TypeError, ValueError):
        inliner.reporter.warning(
            'unable to expand %s extlink with base URL %r, please make '
            'sure the base contains \'%%s\' exactly once'
            % (typ, base_url), line=lineno)
        full_url = base_url + part
    if not has_explicit_title:
        if prefix is None:
            title = full_url
        else:
            title = prefix + part
    pnode = nodes.reference(title, title, internal=False, refuri=full_url)
    return [pnode], []


def command_parse(env, sig, signode):
    # x, y = sig.split()
    signode += addnodes.literal_emphasis(sig, sig)
    # signode += addnodes.literal_strong(sig, sig)  # needs Sphinx >= 1.3
    return sig

# import re
# from docutils.nodes import Node, TextElement, system_message
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Type



class CountRole(SphinxRole):
    # amp_re = re.compile(r'(?<!&)&(?![&\s])')
    # counters = {}  # class attribute. same dict for all instances.

    def run(self) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
        counters = self.env.temp_data.setdefault('_atelier_counters', {})
        num = counters.get(self.rawtext, None)
        if num is None:
            num = counters[self.rawtext] = 1
        else:
            num += 1
            counters[self.rawtext] = num
        txt = self.format_number(num)
        node = nodes.inline(rawtext=txt, classes=[self.name])
        node += nodes.Text(txt)
        return [node], []

    def format_number(self, num):
        return str(num)


def html_page_context(app, pagename, templatename, context, doctree):
    # experimental. no result yet.
    # print(20151006, pagename, context.keys())
    if pagename.startswith('api/') and pagename != "api/index":
        modname = pagename[4:]
        mod = import_module(modname)
        s = srcref(mod)
        if s:
            tpl = """<p align="right"><a href="{0}">[source]</a></p>"""
            context.update(source_code_link=tpl.format(s))
    #     else:
    #         context.update(
    #             source_code_link="no source code for {0}".format(modname))
    # else:
    #     context.update(
    #         source_code_link="no module in {0}".format(pagename))


def setup(app):
    def add(**kw):
        # skw = dict()
        # for k, v in kw.items():
        #     skw[str(k)] = str(v)
        #
        app.add_object_type(**kw)

    add(directivename='management_command',
        rolename='manage',
        indextemplate='pair: %s; management command')

    # add(directivename='role', rolename='role',
    #     indextemplate='pair: %s; docutils role')
    # add(directivename='directive', rolename='directive',
    #     indextemplate='pair: %s; docutils directive')

    # add(directivename='fab_command',
    #     rolename='fab',
    #     indextemplate='pair: %s; fab command')

    app.add_object_type(
        'command', 'cmd', 'pair: %s; command', command_parse)

    add(directivename='xfile',
        rolename='xfile',
        indextemplate='pair: %s; file')
    add(directivename='setting', rolename='setting',
        indextemplate='pair: %s; setting')
    # add(directivename='fixture', rolename='fixture',
    #     objname='demo fixture name',
    #     indextemplate='pair: %s; demo fixture name')
    # add(directivename='actorattr', rolename='aa',
    #     indextemplate='pair: %s; actor attribute')
    add(directivename='screenshot', rolename='screen',
        indextemplate='pair: %s; screenshot')
    # removed 20210426
    # add(directivename='modattr', rolename='modattr',
    #     indextemplate='pair: %s; model attribute')
    # add(directivename='model',
    #     rolename='model', indextemplate='pair: %s; model')
    # app.connect('build-finished', handle_finished)

    # removed 20220528. Not sure whether they are still needed. 
    # if 'name' in atelier.current_project.SETUP_INFO:
    #
    #     app.connect('autodoc-skip-member', autodoc_skip_member)
    #     app.connect('autodoc-process-docstring', autodoc_add_srcref)

    # app.connect(str('autodoc-process-signature'), process_signature)
    app.connect('html-page-context', html_page_context)

    app.add_role('coderef', coderef_role)
    app.add_role('message', message_role)

    # 20210426 app.add_role(str('actor'), actor_role)

    roles.register_canonical_role('blogref', blogref_role)

    roles.register_local_role('count', CountRole())

    app.add_config_value(
        'blogref_format',
        "https://luc.lino-framework.org/blog/%Y/%m%d.html", 'html')



# -*- coding: UTF-8 -*-

from atelier.test import TestCase

class BasicTests(TestCase):

    def test_readme(self):
        self.run_simple_doctests('README.rst')
    def test_rstgen(self):
        self.run_simple_doctests('rstgen/__init__.py')
    def test_rstgen(self):
        self.run_simple_doctests('rstgen/utils.py')

class SphinxTests(TestCase):
    def test_sphinxconf(self):
        self.run_simple_doctests('rstgen/sphinxconf/__init__.py')

    def test_base(self):
        self.run_simple_doctests('rstgen/sphinxconf/base.py')

    def test_sigal(self):
        self.run_simple_doctests('rstgen/sphinxconf/sigal_image.py')



# -*- coding: utf-8 -*-
# Copyright 2011-2020 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

"""
Defines the :func:`rstgen.sphinxconf.interproject.configure` function.

"""
import os

from pathlib import Path
# from importlib import import_module
from sphinx.util import logging ; logger = logging.getLogger(__name__)

# from invoke import Context

# import atelier

USE_LOCAL_BUILDS = os.environ.get("ATELIER_IGNORE_LOCAL_BUILDS", "") != "yes"

# Whether to use objects.inv files from other local doctrees if they exist.
# E.g. on Travis no other projects are installed from source, so there we
# cannot use it.


def configure(globals_dict, prjspec=None, **nicknames):
    """

    Install doctrees of all (or some) atelier projects into the
    :envvar:`intersphinx_mapping` of your :xfile:`conf.py`.

    See :doc:`/sphinxext/interproject`.

    """

    intersphinx_mapping = globals_dict.setdefault('intersphinx_mapping', dict())
    # extlinks = dict()

    # this = atelier.current_project
    # if this is None:
    #     raise Exception("current_project in {} is None!".format(globals_dict['__file__']))

    this_conf_file = Path(globals_dict['__file__']).resolve()

    if prjspec:
        from atelier.projects import get_project_from_module
        if isinstance(prjspec, str):
            prjspec = prjspec.split()
        prjlist = [get_project_from_module(n) for n in prjspec]

    else:
        prjlist = []
        # 20210716 feature removed because it causes unreproduceable situation
        # raise Exception("20210716")
        # for p in load_projects():
        # for p in reversed(list(load_projects())):
        #     if not str(this_conf_file).startswith(str(p.root_dir)):
        #         # print("20190122 {} startswith  {}".format(this_conf_file, p.root_dir))
        #         prjlist.append(p)

    if nicknames:
        from atelier.projects import get_project_from_nickname
        for k, v in nicknames.items():
            p = get_project_from_nickname(k)
            if p:
                # makedocs in demo projects in the book failed to load book
                # if not str(this_conf_file).startswith(str(p.root_dir)):
                if this_conf_file.parent != p.root_dir:
                    prjlist.append(p)
                # else:
                #     print("20210708 skipping {} because {}".format(p, this_conf_file))
            else:
                intersphinx_mapping[k] = v

    # logger.info("20180907 prjlist {}".format(prjlist))
    for prj in prjlist:
        # This will load the `tasks.py` of other
        # projects. Possible side effects.
        # print("20180428 {} {}".format(prj.name, prj.config['doc_trees']))
        # config = prj.inv_namespace.configuration()
        # print("20210716 {}".format(prj))
        # ctx = Context(config)
        # for doc_tree in prj.config['doc_trees']:
        count = 0
        for doc_tree in prj.get_doc_trees():
            if not doc_tree.has_intersphinx:
                logger.info("%s has no intersphinx", p)
                continue
            count += 1
            urls = prj.get_xconfig('intersphinx_urls', {})
            url = urls.get(doc_tree.rel_path)
            if not url:
                if prjspec:
                    logger.warning(
                        "No intersphinx mapping for {} of {} ({})".format(
                            doc_tree.rel_path, prj.nickname, urls))
                continue

            # if prj.nickname == "getlino":
            #     raise Exception("20191003 {}".format(doc_tree.src_path))

            p = None
            src_path = doc_tree.src_path
            if src_path is not None:
                if this_conf_file == src_path / 'conf.py':
                    # don't add myself to intersphinx.
                    continue

                if USE_LOCAL_BUILDS:
                        # print("20190306a", doc_tree, src_path)
                    # p = prj.root_dir / (doc_tree + '/.build/objects.inv')
                    p = src_path / '.build/objects.inv'
                    if p.exists():
                        pass
                        # logger.info("Found local {}".format(p))
                    else:
                        logger.info("File %s does not exist", p)
                        p = None


            # The unique identifier can be used to prefix cross-reference targets
            # http://www.sphinx-doc.org/en/master/ext/intersphinx.html#confval-intersphinx_mapping
            k = prj.nickname + doc_tree.rel_path
            k = k.replace('_', '')
            k = str(k)

            if k in intersphinx_mapping:
                logger.info("Modify intersphinx key %s from %s to %s",
                    k, intersphinx_mapping[k], p)
            #     raise Exception("Duplicate intersphinx key {} used for {} "
            #                     "(you ask to redefine it to {})".format(
            #         k, intersphinx_mapping[k], p))
            if p is not None:
                p = str(p)
            intersphinx_mapping[k] = (url, p)

        if count == 0 and prjspec:
            logger.warning("No doctree for {}".format(prj))
        # if prj.srcref_url:
        #     k = '%s_srcref' % prj.nickname
        #     extlinks[str(k)] = (prj.srcref_url, '')

    # atelier.current_project = this
    globals_dict.update(intersphinx_mapping=intersphinx_mapping)

    # logger.info("20190306 prjlist is {}, intersphinx_mapping is {}".format(
    #     prjlist, intersphinx_mapping))

    # if False:  # no longer used
    #     globals_dict.update(extlinks=extlinks)



# -*- coding: utf-8 -*-
# Copyright 2011-2022 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
# docs: https://atelier.lino-framework.org/rstgen.html

"""See :doc:`/rstgen`.

.. autosummary::
   :toctree:

   utils
   sphinxconf


"""

import sys
import io

SETUP_INFO = {'url': 'https://gitlab.com/lino-framework/rstgen'}

# can be set from your conf.py
# used e.g. in default_conf.py
_config = {
    'public_url': '',
    'selectable_languages': '',
    'use_dirhtml': False,
}

def set_config_var(**kwargs):
    for k, v in kwargs.items():
        if not k in _config:
            raise Exception("Invalid config var {}".format(k))
        _config[k] = v
    # print("20220529 set {}".format(_config))
    # testing e.g. in lucet

def get_config_var(k):
    # if k == "selectable_languages":
    #     # raise Exception()
    #     print("20220529 get_config({}) --> {}".format(k, _config[k]))
    return _config[k]


class Column(object):
    # A column in a Table.

    def __init__(self, table, index, header, width=None):
        self.table = table
        self.header = header
        self.width = width
        self.index = index

    def __str__(self):
        return "{}[{}]:{}".format(self.table, self.index, self.header)

    def adjust_width(self, row):
        # Adjust required width of this column for the given row.
        s = self.table.format_value(row[self.index])
        s = str(s)
        for ln in s.splitlines():
            if self.width is None or self.width < len(ln):
                self.width = len(ln)
                # if self.width > 500:
                #     raise Exception("width %r more than 500" % s)


def header(level, text):
    result = io.StringIO()

    def writeln(s=''):
        result.write(s + '\n')

    _write_header(writeln, level, text)
    return result.getvalue()


def write_header(fd, level, text):
    # Write the string `text` to file `fd` as a header of the given `level`.

    def writeln(text=''):
        fd.write(text + '\n')

    _write_header(writeln, level, text)


def _write_header(writeln, level, s):
    if level == 1:
        writeln('=' * len(s))
    elif level == 2:
        writeln('-' * len(s))
    elif level == 3:
        writeln('~' * len(s))
    writeln(s)
    if level == 1:
        writeln('=' * len(s))
    elif level == 2:
        writeln('-' * len(s))
    elif level == 3:
        writeln('~' * len(s))
    elif level == 4:
        writeln('=' * len(s))
    elif level == 5:
        writeln('-' * len(s))
    elif level == 6:
        writeln('~' * len(s))
    else:
        raise Exception("Invalid level %d" % level)
    writeln()


class Table(object):
    simple = True

    def format_value(self, v):
        return str(v)

    def __init__(self, headers, show_headers=True):
        self.headers = [self.format_value(h) for h in headers]
        self.show_headers = show_headers
        self.cols = [Column(self, i, h) for i, h in enumerate(headers)]
        self.adjust_widths(headers)

    def adjust_widths(self, row):
        if len(row) != len(self.headers):
            raise Exception("Invalid row %(row)s" % dict(row=row))
        for col in self.cols:
            col.adjust_width(row)
            if '\n' in row[col.index]:
                self.simple = False

    def format_row(self, row):
        # ~ return ' '.join([unicode(row[c.index]).ljust(c.width) for c in self.cols])
        lines = [[] for x in self.cols]
        for c in self.cols:
            cell = row[c.index]
            for ln in cell.splitlines():
                lines[c.index].append(ln.ljust(c.width))
        height = 1
        for c in self.cols:
            height = max(height, len(lines[c.index]))
        for c in self.cols:
            while len(lines[c.index]) < height:
                lines[c.index].append(''.ljust(c.width))
        x = []
        for i in range(height):
            x.append(self.margin
                     +
                     self.colsep.join(
                         [' ' + lines[c.index][i] + ' ' for c in self.cols])
                     + self.margin)
        return '\n'.join(x)

    def write(self, fd, data=[]):
        # ~ if len(data) == 0:
        # ~ raise Exception("No rows in %r" % data)
        rows = []
        for i, row in enumerate(data):
            assert len(row) == len(self.cols)
            rows.append([self.format_value(v) for v in row])

        for row in rows:
            self.adjust_widths(row)

        for c in self.cols:
            if c.width is None:
                c.width = 1
                # raise Exception("width {} is None".format(c))

        if self.simple:
            self.header1 = ' '.join([('=' * (c.width + 2)) for c in self.cols])
            self.header2 = ' '.join([('-' * (c.width + 2)) for c in self.cols])
            self.margin = ''
            self.colsep = ' '
        else:
            self.header1 = '+' + \
                           '+'.join([('-' * (c.width + 2)) for c in self.cols]) + '+'
            self.header2 = '+' + \
                           '+'.join([('=' * (c.width + 2)) for c in self.cols]) + '+'
            self.margin = '|'
            self.colsep = '|'

        def writeln(s):
            fd.write(s.rstrip() + '\n')

        writeln(self.header1)
        if self.show_headers:
            writeln(self.format_row(self.headers))
            writeln(self.header2)
        for row in rows:
            writeln(self.format_row(row))
            if not self.simple:
                writeln(self.header1)
        if self.simple:
            writeln(self.header1)

    def to_rst(self, rows):
        if len(rows) == 0:
            return "\n"
        fd = io.StringIO()
        self.write(fd, rows)
        return fd.getvalue()


def table(headers, rows=tuple(), **kw):
    t = Table(headers, **kw)
    return t.to_rst(rows)


def ul(items, bullet="-"):
    s = ""
    compressed = True
    for i in items:
        if '\n\n' in i:
            compressed = False
            break

    innersep = '\n' + (' ' * (len(bullet) + 1))
    if compressed:
        for i in items:
            text = innersep.join(i.splitlines())
            s += "%s %s\n" % (bullet, text)
    else:
        for i in items:
            text = innersep.join(i.splitlines())
            s += "\n%s %s\n" % (bullet, text)
    return s


def ol(items, bullet="#."):
    return ul(items, bullet)


def boldheader(title):
    return "\n\n**%s**\n\n" % str(title).strip()


#
def toctree(*children, **options):
    text = "\n\n.. toctree::"
    for k, v in options.items():
        text += "\n    "
        text += ":{0}:".format(k)
        if isinstance(v, str):
            text += " " + v
        elif v is True:
            pass
        else:
            text += " " + str(v)

    text += "\n"
    for child in children:
        text += "\n    " + child
    text += "\n"
    return str(text)


class stdout_prefix(object):
    # experimental
    def __init__(self, prefix):
        self.prefix = '\n' + prefix
        self.saved_stdout = sys.stdout

    def __enter__(self):
        sys.stdout = self

    def write(self, s):
        s = self.prefix + self.prefix.join(s.splitlines())
        self.saved_stdout.write(s)

    def __exit__(self, type, value, traceback):
        sys.stdout = self.saved_stdout
        self.writer = None


def attrtable(rows, cols):
    if isinstance(cols, str):
        cols = cols.split()
    cells = [[str(getattr(obj, k)) for k in cols] for obj in rows]
    return table(cols, cells)



# -*- coding: utf-8 -*-
# this is being read using exec() by rstgen.sphinxconf.configure()

from pathlib import Path
import rstgen

# from rstgen.sphinxconf import configure ; configure(globals())

docs_root = Path(__file__).parent.absolute()
# print("20210426", docs_root)

# prj = atelier.current_project
# if prj.main_package:
#     root_mod, tpl = srcref_url_template(prj.main_package)
#     if tpl:
#         extlinks.update(srcref=(tpl, ''))

use_dirhtml = rstgen.get_config_var('use_dirhtml')
selectable_languages = rstgen.get_config_var('selectable_languages')

public_url = html_context.setdefault(
    'public_url', rstgen.get_config_var('public_url'))
    # 'public_url', prj.get_public_docs_url('docs'))

def lngfmt(lng):
    return lng.upper()

def get_page_url(language, pagename):
    if use_dirhtml:
        if pagename == master_doc:
            pagename = ""
        elif pagename.endswith("/" + master_doc):
            pagename = pagename[:-len(master_doc)]
    else:
        pagename += ".html"
    if not selectable_languages:
        return pagename
    if language == selectable_languages[0]:
        return pagename
    return language + "/" + pagename

def get_lang_selector(language, pagename, sep=" | "):
    # public_url = prj.get_xconfig('intersphinx_urls')['docs']
    # public_url = prj.get_public_docs_url('docs')
    # public_url = html_context['public_url']
    if not selectable_languages:
        return ''
    if len(selectable_languages) <= 1:
        return ''
    # get_page_url()
    filename = pagename + source_suffix
    if use_dirhtml:
        homename = ""
    else:
        homename = master_doc + source_suffix

    html = ''
    for i, lng in enumerate(selectable_languages):
        if i > 0:
            html += sep
        if lng == language:
            html += "<b>{}</b>".format(lngfmt(lng))
        else:
            href = public_url + "/" + get_page_url(lng, pagename)
            if i == 0:
                pth = docs_root.parent / "docs" / filename
                if not pth.exists():
                    href = homename
            else:
                pth = docs_root.parent / (lng + "docs") / filename
                if not pth.exists():
                    href = lng + "/" + homename
                    # print("Not found b:", pth)
            html += '<a href="{}">{}</a>'.format(href, lngfmt(lng))
    # html += "</p>"
    return html

html_context.update(get_lang_selector=get_lang_selector)
html_context.update(get_page_url=get_page_url)


# extlinks = {}
# extensions = []
# templates_path = []



# extensions += ['sphinxcontrib.youtube']
extensions += ['rstgen.sphinxconf.blog']
extensions += ['rstgen.sphinxconf.complex_tables']
# extensions += ['lino.sphinxcontrib.actordoc']
# extensions += ['sphinxcontrib.taglist']
# extensions += ['sphinx.ext.inheritance_diagram']

# extensions += ['rstgen.sphinxconf.sigal_image']
# sigal_base_url = 'https://sigal.saffre-rumma.net'

# extensions += ['sphinx_panels'] # pip install sphinx-panels
extensions += ['sphinx_design'] # pip install sphinx-design

# source_suffix = '.rst'
# master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
language = 'en'

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%d.%B.%Y'

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = 'default'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = "⌘"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = "logo.jpg"

# same logo is being used for all languages
# html_logo = str(Path().absolute() / '../docs/.static/rumma100.png') # pandas

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
pth = docs_root / '.static'
if (pth).exists():
    html_static_path.append(str(pth))

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%Y-%m-%d'
#~ html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True


# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_use_modindex = False

# If false, no index is generated.
html_use_index = False

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, the reST sources are included in the HTML build as _sources/<name>.
html_copy_source = False

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''
# html_use_opensearch = lino.SETUP_INFO['url']

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = ''

# Output file base name for HTML help builder.
# htmlhelp_basename = 'hg'

html_css_files = ['atelier.css']

if html_theme == "sphinx_rtd_theme":
    html_theme_options = {
        "prev_next_buttons_location": "both",
        "style_nav_header_background": "#dddddd",
        "style_external_links": False,  # disadvantage: line spacing increases for lines with a link
        "includehidden": False,
    }
    html_sidebars = {
        # '**': ['select_lang.html', 'searchbox.html'],
        '**': ['languages.html', 'globaltoc.html', 'searchbox.html', 'links.html'],
    }
    extensions.append('sphinx_rtd_theme')

elif html_theme == "alabaster":
    my_font_family = "Swiss, Helvetica, 'Liberation Sans'"
    html_theme_options={
        "font_family": my_font_family,
        "head_font_family": my_font_family,
    }
    html_sidebars = {
        # '**': ['select_lang.html', 'searchbox.html'],
        '**': ['globaltoc.html', 'searchbox.html', 'links.html'],
    }
elif html_theme == "pydata_sphinx_theme":
    # https://pydata-sphinx-theme.readthedocs.io/en/latest/user_guide/index.html
    html_theme_options = {
        # "gitlab_url": "https://gitlab.com/saffre-rumma/www",
        "globaltoc_includehidden": True,
        # "collapse_navigation": False,
        "navigation_with_keys": True,
        # "navigation_depth": 2,
        "show_prev_next": False,
        "search_bar_text": "Search...",
        "navbar_center": ["search-field.html"],
        "navbar_end": ["languages.html", "navbar-icon-links.html"],
    }
    panels_add_bootstrap_css = False
    # html_css_files += ["https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css"]
    html_sidebars = {
      '**':  [
        # 'languages.html',
        # 'page-toc.html',
        # 'globaltoc.html',
        # 'sidebar-search-bs.html' ,
        'sidebar-nav-bs.html'
        ],
    }
elif html_theme == "insipid":
    html_css_files.append('insipid-cust.css')
    html_theme_options = {
        # 'body_max_width': None,
        # 'breadcrumbs': True,
        'globaltoc_includehidden': False,
        'left_buttons': [
            'search-button.html',
            'home-button.html',
        ],
        'right_buttons': [
            'fullscreen-button.html',
            'repo-button.html',
            'facebook-button.html',
        ],
    }
    html_sidebars = {
        '**' : ['languages.html',
            'globaltoc.html',
            'separator.html',
            'searchbox.html',
            'indices.html',
            'links.html'
            ]}

# intersphinx_mapping = dict(
#     cg=('https://community.lino-framework.org/', None))



# -*- coding: utf-8 -*-
# Copyright 2011-2022 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

"""
Sphinx extensions and a :func:`configure` function used to build
the documentation trees maintained by us.

.. autosummary::
   :toctree:

   base
   dirtables
   refstothis
   insert_input
   sigal_image
   complex_tables
   blog
   interproject
"""

import logging ; logger = logging.getLogger(__name__)

import sys
import os

from pathlib import Path
from pkg_resources import parse_version
import sphinx

def configure(globals_dict, project=None, conffiles=["default_conf.py"]):
    """
    Adds to your `conf.py` an arbitrary series of things that all our
    Sphinx docs configuration files have in common.

    Usage: add the following two lines at the beginning of your
    :xfile:`conf.py`::

      from rstgen.sphinxconf import configure
      configure(globals())

    Incomplete list of settings that will be set:

    - `templates_path`
    - `extensions`, `extlinks` and `intersphinx_mapping` are initialized but empty
    - `master_doc` = 'index'
    - `source_suffix` = '.rst'
    - `primary_domain` = 'py'
    - `pygments_style` = 'sphinx'
    - `version` and `release` from SETUP_INFO
    - `exclude_trees` = ['old', 'include', '.build']

    """
    docs_root = Path(globals_dict['__file__']).parent.absolute()
    sys.path.append(docs_root)

    # https://insipid-sphinx-theme.readthedocs.io/en/0.2.0/configuration.html
    globals_dict.setdefault('html_theme', "insipid")
    # globals_dict.setdefault('html_theme', "pydata_sphinx_theme")
    # globals_dict.update(html_theme="bizstyle")

    globals_dict.setdefault('html_context', dict())
    globals_dict.setdefault('extlinks', dict())
    globals_dict.setdefault('intersphinx_mapping', dict())
    globals_dict.update(exclude_patterns=['old', 'include', 'shared', '.build'])
    extensions = []
    globals_dict.update(extensions=extensions)

    extensions += [
        'sphinx.ext.inheritance_diagram',
        'sphinx.ext.todo',
        'sphinx.ext.extlinks',
        'sphinx.ext.graphviz',
        'sphinx.ext.intersphinx',
        # no i18n, no discovery, only one entry per doc,
        # 'sphinxcontrib.newsfeed',
        #~ 'sphinx.ext.doctest',
        'rstgen.sphinxconf.base',
        'rstgen.sphinxconf.dirtables',
        'rstgen.sphinxconf.refstothis',
        'rstgen.sphinxconf.insert_input',
    ]

    # default config for autosummary:
    globals_dict.update(autosummary_generate=True)
    if parse_version(sphinx.__version__) < parse_version("1.8"):
        globals_dict.update(autodoc_default_flags=[
            'show-inheritance', 'members'])

    else:
        globals_dict.update(autodoc_default_options={
            'members': None, 'show-inheritance': None})

    mydir = Path(__file__).parent.absolute()
    tp = globals_dict.setdefault('templates_path', [])
    # tp += ['.templates', str(mydir)]
    tp.append(str(mydir / 'templates'))

    globals_dict.setdefault('html_static_path', [str(mydir / 'static')])

    # Some settings I use in all projects:

    globals_dict.update(master_doc='index')
    globals_dict.update(source_suffix='.rst')
    globals_dict.update(primary_domain='py')
    globals_dict.update(pygments_style='sphinx')
    globals_dict.update(autodoc_member_order='bysource')
    globals_dict.update(autodoc_inherit_docstrings=False)

    for fn in conffiles:
        if not os.path.sep in fn:
            fn = mydir / fn
        with open(fn, "rb") as fd:
            exec(compile(fd.read(), fn, 'exec'), globals_dict)

def version2rst(self, m):
    """
    used in docs/released/index.rst
    """
    v = m.__version__
    if v.endswith('+'):
        v = v[:-1]
        print("The current stable release is :doc:`%s`." % v)
        print("We are working on a future version in the code repository.")
    elif v.endswith('pre'):
        print("We're currently working on :doc:`%s`." % v[:-3])
    else:
        print("The current stable release is :doc:`%s`." % v)
        #~ print("We're currently working on :doc:`coming`.")



# -*- coding: utf-8 -*-
# Copyright 2011-2021 by Rumma & Ko Ltd.
# License: GNU Affero General Public License v3 (see file COPYING for details)

"""Defines the :rst:dir:`textimage` and :rst:dir:`complextable`
directives.

"""
from pathlib import Path
from docutils.parsers.rst import directives

import rstgen

from .insert_input import InsertInputDirective


class TextImageDirective(InsertInputDirective):
    """Defines the :rst:dir:`textimage` directive."""

    required_arguments = 1
    final_argument_whitespace = True
    option_spec = dict(scale=directives.unchanged)
    #~ optional_arguments = 4

    def get_rst(self):
        #~ print 'MainBlogIndexDirective.get_rst()'
        #~ env = self.state.document.settings.env
        #~ print self.arguments, self.options, self.content
        left = '\n'.join(self.content)
        right = ''
        for arg in self.arguments[0].split():
            right += '.. figure:: %s\n' % arg
            for i in list(self.options.items()):
                right += "  :%s: %s\n" % i
            right += "\n  %s\n\n" % arg
            #~ right += "\n  \n\n" % arg

        return rstgen.table(["", ""], [[left, right]], show_headers=False)


class ComplexTableDirective(InsertInputDirective):
    """Defines the :rst:dir:`complextable` directive."""

    required_arguments = 0
    final_argument_whitespace = True
    option_spec = dict(header=directives.flag)
    #~ option_spec = dict(scale=unchanged)
    #~ optional_arguments = 4

    def get_rst(self):
        #~ print 'MainBlogIndexDirective.get_rst()'
        #~ env = self.state.document.settings.env
        #~ print self.arguments, self.options, self.content
        cellsep = '<NEXTCELL>'
        rowsep = '<NEXTROW>'
        if len(self.arguments) > 0:
            cellsep = self.arguments[0]
        if len(self.arguments) > 1:
            rowsep = self.arguments[1]

        content = '\n'.join(self.content)
        rows = []
        colcount = None

        for row in content.split(rowsep):
            cells = [cell.strip() for cell in row.split(cellsep)]
            if colcount is None:
                colcount = len(cells)
            else:
                assert colcount == len(cells)
            rows.append(cells)

        if 'header' in self.options:
            return rstgen.table(rows[0], rows[1:])

        return rstgen.table([""] * colcount, rows, show_headers=False)


container_tpl = """

.. raw:: html

    <div class="container-fluid">
        {% for col_class, row in rows %}
        <div class="row">
           {% for cell_content in row %}
            <div class="{{col_class}}">
                <div class="card text-center intro-card shadow">
                <div class="card-body flex-fill">

{{cell_content}}

.. raw:: html

                </div>
                </div>
            </div>

           {% endfor %}

.. raw:: html

        </div>

        {% endfor %}

.. raw:: html

    </div>

    """

card_class_map = {
    2 : "col-md-6 col-sm-6 col-xs-12 d-flex align-items-start",
    3 : "col-md-4 col-sm-6 col-xs-12 d-flex align-items-start",
    4 : "col-md-3 col-sm-6 col-xs-12 d-flex align-items-start",
}

from jinja2 import Environment
JINJA_ENV = Environment()


class CardsDirective(InsertInputDirective):
    """Defines the :rst:dir:`cards` directive."""

    # titles_allowed = True
    required_arguments = 0
    final_argument_whitespace = True

    def get_rst(self):
        cellsep = '<NEXTCARD>'
        rowsep = '<NEXTROW>'

        rows = []
        content = '\n'.join(self.content)
        for row in content.split(rowsep):
            cells = [cell.strip() for cell in row.split(cellsep)]
            col_class = card_class_map[len(cells)]
            rows.append([col_class, cells])

        # print(rows)
        tpl = JINJA_ENV.from_string(container_tpl)
        s = tpl.render(rows=rows)
        # pth = Path("~/tmp/tmp.rst").expanduser()
        # with open(pth, 'w') as fh:
        #     fh.write(s)
        return s


def setup(app):
    app.add_directive('textimage', TextImageDirective)
    app.add_directive('complextable', ComplexTableDirective)
    app.add_directive('cards', CardsDirective)



# -*- coding: utf-8 -*-
# Copyright 2016-2021 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

"""A Sphinx extension that adds the :rst:dir:`refstothis` directive.

Thanks to Tomasz Czyż who inspired me to write this.  His
`sphinxcontrib-taglist
<https://github.com/spinus/sphinxcontrib-taglist>`_ module uses a
different approach, based on the idea that every "reference" to a tag
also specifies a text to appear in the list of references.
:rst:dir:`refstothis` is for people who don't want to specify their
own text for each tag directive.

TODO:

- When I do :cmd:`inv clean` followed by :cmd:`inv bd`, then e.g. the
  `refstothis` in :doc:`/refstothis/baz` does not mention all other documents.
  But when I then touch the `baz.rst` file and run :cmd:`inv bd`, then they are
  all there.  I guess this is because not all documents have been loaded when
  the rst is being generated.

"""


from os import path
import time
from pprint import pprint
from sphinx import addnodes
from sphinx.util import docname_join
import rstgen

from .insert_input import InsertInputDirective


def py2rst(x):
    x = pprint(x)
    # x  = x.replace("|","")
    return x


class RefsToThis(InsertInputDirective):
    """Implements the rst:dir:`refstothis` directive."""

    # debug = True

    def get_rst(self):
        # return str(self.state.document.refnames)
        env = self.state.document.settings.env
        # print("20211108", self.content)
        targets = set(self.content)
        if len(targets) == 0:
            targets = {env.temp_data['docname']}
        # target = ' '.join(self.content).strip()
        # if not target:
        #     # print("20140409 target is %r" % target)
        found = set()
        rows = set()

        # headers = 'children attributes \
        # resolved referenced indirect_reference_name \
        # tagname'.split()

        for docname in env.found_docs:
            if env.temp_data['docname'] == docname:  # skip myself
                continue

            try:
                doc = env.get_doctree(docname)
            except Exception:
                # 20140117 i had the following after a fab clean:
                #   File "/home/luc/pythonenvs/py27/local/lib/python2.7/site-packages/sphinx/environment.py", line 1077, in get_doctree
                #     f = open(doctree_filename, 'rb')
                # IOError: [Errno 2] No such file or directory: u'/home/luc/hgwork/lino/docs/.build/.doctrees/topics/names.doctree'

                continue

            # print("20140115 traversing", docname)
            for ref in doc.traverse(addnodes.pending_xref):
                if ref['reftype'] == 'doc':
                    other = docname_join(ref['refdoc'], ref['reftarget'])
                else:
                    other = ref['reftarget']
                if other in targets:
                    found.add(ref['refdoc'])
                    # print("20140409 found", ref)
                else:
                    # rows.add(ref['reftarget'])
                    rows.add(other)
                    # rows.add(repr(ref.attributes))
                    # row = []
                    # for h in headers:
                    #     row.append(py2rst(getattr(ref, h, 'N/A')))
                    # rows.append(unicode(row))

        if len(found) == 0:
            s = """No documents found for targets %r.""" % targets
            # s += """\nPending xrefs were %r.""" % rows
            return s

        entries = []
        for refdoc in found:
            mtime = path.getmtime(env.doc2path(refdoc))
            entries.append((mtime, refdoc))

        def f(a):
            return a[0]
        entries.sort(key=f)
        entries.reverse()

        # from time import strftime

        items = [':doc:`/%(doc)s` (%(time)s)' % dict(
            time=time.ctime(e[0]),
            doc=e[1]) for e in entries]

        if 'debug' in self.options:
            items.append("DEBUG: pending xrefs were %r." % rows)

        return rstgen.ul(items)


def setup(app):
    app.add_directive('refstothis', RefsToThis)



from setuptools import setup

SETUP_INFO = dict(
    name='rstgen',
    version='22.12.0',
    packages=['rstgen', 'rstgen.sphinxconf', 'rstgen.sphinxconf.languages'],
    install_requires=[],
    tests_require=['atelier'],
    test_suite='tests',
    description="Pythonic API for generating reStructuredText.",
    license_files=['COPYING'],
    url='https://gitlab.com/lino-framework/rstgen',
    author='Rumma & Ko Ltd',
    author_email='info@lino-framework.org')

SETUP_INFO.update(classifiers="""\
Programming Language :: Python
Programming Language :: Python :: 3
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
License :: OSI Approved :: GNU Affero General Public License v3
Natural Language :: English
Operating System :: OS Independent""".splitlines())

SETUP_INFO.update(long_description=open("README.rst").read())

SETUP_INFO.update(
    zip_safe=False,
    include_package_data=True)



if __name__ == '__main__':
    setup(**SETUP_INFO)



# -*- coding: utf-8 -*-
# Copyright 2016-2019 Rumma & Ko Ltd.
# License: GNU Affero General Public License v3 (see file COPYING for details)

"""Defines the :rst:dir:`directory`, :rst:dir:`tickets_table` and
:rst:dir:`entry_intro` directives.

.. rst:directive:: directory

Inserts a table containing three columns 'title', 'author' and 'date',
and one row for each `.rst` file found in this directory (except for
the calling file).

.. rst:directive:: tickets_table

This was used e.g. to build http://lino-framework.org/tickets until 20190329
(where we threw it away as it failed in Sphinx 2)

.. rst:directive:: entry_intro

This doesn't yet work unfortunately. Was here until 20190329 (where we threw it
away as it failed in Sphinx 2).

"""

from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
from builtins import filter
from builtins import object

import logging ; logger = logging.getLogger(__name__)

from os.path import abspath, dirname, join

from docutils.parsers.rst import directives


from sphinx.util import docname_join
from sphinx.util.matching import patfilter

import rstgen

from .insert_input import InsertInputDirective

package_dir = abspath(dirname(__file__))

from jinja2 import FileSystemLoader, TemplateNotFound
from jinja2 import Environment
# from jinja2.sandbox import SandboxedEnvironment as Environment

template_dirs = [join(package_dir, 'templates')]
template_loader = FileSystemLoader(template_dirs)
template_env = Environment(loader=template_loader)


def render_entry(tplname, context):
    try:
        template = template_env.get_template(tplname)
    except TemplateNotFound:
        template = template_env.get_template('dirtables/entry.rst')
    return template.render(context)


def rel_docname(a, b):
    """
    >>> print(rel_docname('tickets/index','tickets/2'))
    2
    >>> print(rel_docname('tickets/index','/todo/index'))
    /todo/index
    """
    if b.startswith('/'):
        return b
    a1 = a.rsplit('/')[0] + '/'
    if b.startswith(a1):
        return b[len(a1):]
    return b


class Entry(object):

    def __init__(self, docname, title, meta):
        self.docname = docname
        self.title = title
        self.meta = meta

    @classmethod
    def create(cls, env, docname):
        return cls(rel_docname(env.docname, docname),
                   env.titles.get(docname),
                   env.metadata.get(docname))


class DirectoryTable(InsertInputDirective):
    entry_class = Entry
    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {
        'filter': directives.unchanged,
        'orderby': directives.unchanged,
    }

    def get_rst(self):
        env = self.state.document.settings.env
        entries = []
        all_docnames = env.found_docs.copy()
        found = set([env.docname])  # don't include myself
        for entry in self.content:
            if not entry:
                continue
            patname = docname_join(env.docname, entry)
            docnames = sorted(patfilter(all_docnames, patname))
            for docname in docnames:
                if not docname in found:
                    found.add(docname)
                    entries.append(self.entry_class.create(env, docname))
        expr = self.options.get('filter')
        if expr:
            def func(e):
                try:
                    return eval(expr, dict(e=e))
                except Exception as exc:
                    return "{} in {}".format(exc, expr)
            entries = list(filter(func, entries))

        orderby = self.options.get('orderby')
        if orderby:
            def func(a):
                return getattr(a, orderby, '')
            entries = sorted(entries, key=func)

        headers = self.get_headers()
        rows = []
        for e in entries:
            rows.append(self.format_entry(e))
        return rstgen.table(headers, rows)

    def get_headers(self):
        return ['title', 'author', 'date']

    def format_entry(self, e):
        cells = []
        # text = ''.join([unicode(c) for c in e.title.children])
        # cells.append(":doc:`%s <%s>`" % (text, e.docname))
        cells.append(":doc:`%s`" % e.docname)
        cells.append(str(e.meta.get('author', '')))
        cells.append(str(e.meta.get('date', '')))
        return cells


# class TicketsTable(DirectoryTable):
#
#     def get_headers(self):
#         return ['title' + ' '*50, 'state', 'module', 'since', 'for']
#
#     def format_entry(self, e):
#         cells = []
#         # cells.append(e.docname)
#         cells.append(":doc:`%s`" % e.docname)
#         # text = ''.join([unicode(c) for c in e.title.children])
#         # cells.append(":doc:`%s <%s>`" % (text, e.docname))
#         cells.append(str(e.meta.get('state', '')))
#         # cells.append(unicode(e.meta.get('reporter', '')))
#         ref = e.meta.get('module', '')
#         if ref:
#             cells.append(":mod:`%s`" % ref)
#         else:
#             cells.append("(N/A)")
#
#         cells.append(str(e.meta.get('since', '')))
#         cells.append(str(e.meta.get('for', '')))
#         return cells
#
#
# class EntryIntro(InsertInputDirective):
#
#     def get_rst(self):
#         env = self.state.document.settings.env
#         docname = env.docname
#         meta = env.process_metadata(docname, self.state.document)
#         # e = Entry.create(env, env.docname)
#         context = dict(this=self,
#                        env=self.state.document.settings.env,
#                        dir=dir,
#                        document=self.state.document,
#                        meta=meta)
#         template = 'dirtables/entry.rst'
#         return render_entry(template, context)

# from docutils import nodes
# from sphinx.roles import XRefRole

# RREFS


# class ReferingRefRole(XRefRole):
#     def result_nodes(self, document, env, node, is_ref):
#         # RREFS[]
#         print("20140115 result_nodes", document, env, node, is_ref)
#         return [node], []


def setup(app):
    app.add_directive('directory', DirectoryTable)
    # app.add_directive('tickets_table', TicketsTable)
    # app.add_directive('entry_intro', EntryIntro)
    # app.add_role(str('rref'), ReferingRefRole(
    #     lowercase=True,
    #     innernodeclass=nodes.emphasis,
    #     warn_dangling=True))

